'' Demo text driver for the ILI9325. Uses fonts stored in local hub ram rather than external ram
'' not much space left for code but this should be useful to get the display working
'' as it is possible to print values on the screen
'' use versions with the font buffered in external ram for faster speed
'' James Moxham 2012

CON


  _clkmode      = xtal1 + pll16x                        ' use crystal x 16
  _xinfreq      = 5_000_000

  _SD_DO = 12
  _SD_CLK = 13
  _SD_DI = 14
  _SD_CS = 15
  _SD_WP = -1                                           ' -1 ifnot installed.
  _SD_CD = -1                                           ' -1 ifnot installed.

  _RTC_DAT = -1                                         ' -1 ifnot installed. For kye's sd driver
  _RTC_CLK = -1                                         ' -1 ifnot installed.

  RamText   = 169984
  VT100     = 202752
  FontTable = 206080 
  
 ' termimal display values
  cols = 40
  rows = 15
  screensize = cols * rows
  lastrow = screensize - cols
  ' TV for debugging
  _TV_Enabled = true                                    ' TV display on
  _TV_Pin = 16                                          ' group of 3 pins starting at pin 16 = 16,17,18

OBJ
 delay:         "timing"                                ' thanks to parallax
 fat:           "SD3.01_FATEngine.spin"                 ' thanks to Kye
 'text:          "tv_text"                               ' thanks to Chip    
 'str:           "ASCII0_STREngine.spin"                 ' thanks to Kye 
 'spi:           "SPI_ASM"                               ' for the touchscreen thanks to Beau
 ILI9325:       "ILI9325_Dracblade"                     ' J Moxham
 'keyboard:      "keyboard"                              ' Chip Gracey

VAR
   'byte LineOfText1[20] ' general purpose string buffer
   'byte LineOfText2[20] ' general purpose string buffer
   byte FontHeight                ' size of the current loaded font (pixels = fontsize *.75)
   word BackFontColor           ' the background color in RRRRRGGG GGGBBBBB format
   word curx                    ' current cursor x position in pixels
   word cury                    ' current cursor y position in pixels
   byte Orientation                ' true for portrait, false for landscape
   word ScreenWidth             ' either 240 in portrait or 320 in landscape
   word ScreenHeight            ' 320 in portrait or 240 in landscape
   byte screen[screensize]
   long col,row,flag
  
PUB Main
    'text.start(_TV_Pin)                                 ' start the TV driver
    'text.clearscreen                                    ' clear the screen
    'PrintStringCR(string("Testing for SD card")) 
    ILI9325.start_ram                                   ' start the external ram driver in a cog
    ILI9325.start_ILI9325                               ' start the ILI9325 display
    SetOrientation(true)                                '  in portrait (true) or landscape mode (false)   
    'keyboard.startx(26, 27, %100, 4)                    'Start Keyboard Driver  pin,pin,num,repeatrate  
    ifnot( fat.FATEngineStart(_SD_DO, _SD_CLK, _SD_DI, _SD_CS, _SD_WP, _SD_CD, _RTC_DAT, _RTC_CLK, -1))   
      reboot
    fat.mountPartition(0)                               ' mount the sd card

    DemoScreen


PUB DemoScreen | c
    SetOrientation(false)                               ' landscape
    NewFont(string("Par16.ifn"))                      ' Parallax 8x16 font (16 point) with no smoothing (small fonts better not smoothed)
    ' if change the font, change the size of the font array at the bottom 
    Text_cls
    Text_str(string("Basic text display for the ILI9325"))
    Text_crlf
    Text_str(string("Print decimal value "))
    Text_dec(100)
    Text_crlf
    text_str(string("Hex value 500 = "))
    Text_hex(500,4)
    text_crlf
    ' add the keyboard object to test the keyboard
      
    'Text_str(string("Type something on the keyboard"))
    'repeat
    '  c :=keyboardtest
    '  if c==13
    '    Text_crlf                  ' otherwise only prints carriage return, not the line feed
    '  else
    '    Text_out(c)                ' send to the display

    
PUB SetOrientation(e) ' change orientation true = portrait. Changes global variable 'orientation' and also screen width and height
   orientation := e
   ILI9325.ChangeOrientation(orientation)
   if orientation
     screenwidth :=239
     screenheight :=319
   else
     screenwidth :=319
     screenheight :=239
{
PUB KeyboardTest
    result:=0
      repeat
        if keyboard.gotkey <>0                          ' scan keyboard
          result:=keyboard.getkey
      until result <>0                                  ' until something comes in 
    if result==200
      result :=8                ' backspace keycode is 200 but convert to 8
}
' Ram version
{
PUB ILILoadFont(stringptr) | filesize,i                ' read a .ifn file premade in angelcode bmfont and then the vb.net program
    ' pass file name in stringptr
    ' the variable "i" must start on 0,256,512,768 etc for the block move to work
    i := fonttable                                      ' store at the constant 'fonttable' location
    fat.openfile(stringptr,"R")                      ' open an image. Size is in the first long and has been rounded up to the next 256 byte block
    filesize := fat.filesize                            ' read the file size. This file has been rounded up to the next 256 byte block
    repeat (filesize >> 8)                              ' read in this number of 256 byte blocks
      fat.readdata(@sdbuffer,256)                       ' get 256 bytes from sd card  
      ILI9325.docmd("A",@sdbuffer,i,256)      ' A must start on 0,256,512,768 etc move data from hub to external ram 256 bytes at a time    
      i += 256                             ' add 256
    fat.closefile
    FontHeight := ILI9325.readrambyte(fonttable+251)    ' extract font size
    ' extract the background color so can draw the background behind the text
    repeat i from 0 to 2
      sdbuffer[i] := ILI9325.readrambyte(fonttable + 35+i)   ' read the 3 RGB bytes
    ILI9325.docmd("E",@sdbuffer,@buffer2,1)             ' convert to 2 byte .ili format, len is number of pixels        
    BackFontColor := word[@buffer2][0]
    result := filesize    
}
PUB ILILoadFont(stringptr)| filesize,i   ' hub version
    ' pass file name in stringptr
    ' the variable "i" must start on 0,256,512,768 etc for the block move to work
    i := 0                                              ' store at the constant 'fonttable' location
    fat.openfile(stringptr,"R")                         ' open an image. Size is in the first long and has been rounded up to the next 256 byte block
    filesize := fat.filesize                            ' read the file size. This file has been rounded up to the next 256 byte block
    repeat (filesize >> 8)                              ' read in this number of 256 byte blocks
      fat.readdata(@font+i,256)                         ' get 256 bytes from sd card
      i += 256                                          ' add 256
    fat.closefile
    FontHeight := font[251]                             ' extract font size
    ' extract the background color so can draw the background behind the text
    repeat i from 0 to 2
      sdbuffer[i] := font[35+i]                         ' read the 3 RGB bytes
    ILI9325.docmd("E",@sdbuffer,@buffer2,1)             ' convert to 2 byte .ili format, len is number of pixels        
    BackFontColor := word[@buffer2][0]                  ' background color in ILI two byte format
    result := filesize
        
{ ' ram version
PUB ILIChar(ascii,x,y) | jump,size,width,height,ramaddress,xoffset,yoffset,xadvance   ' read out a character from a .ifn file stored in ram
  ' pass Fonttable = global
    jump := ILI9325.ReadRamLong((ascii << 2) + Fonttable + 256)' jump location = ascii *4 plus fonttable + 256 
    jump += fonttable                                   ' add the fonttable location
    size := ILI9325.ReadRamLong(jump)                   ' size precalculated = width x height x 2
    xadvance :=  ILI9325.ReadRamLong(jump+20)     '     ' amount to move to next character
    if size > 0 and ascii <> 32                                           ' no need to print anything if size is zero or read more font data
      width :=   ILI9325.ReadRamLong(jump+4)                ' width
      height :=  ILI9325.ReadRamLong(jump+8)            ' height
      xoffset := ILI9325.ReadRamLong(jump+12)           ' xoffset to move
      yoffset := ILI9325.ReadRamLong(jump+16)           ' yoffset to move
      if (x+width -1 + xoffset) > screenwidth
        ILIcrlf ' do a new line if it won't fit
        x :=curx
        y :=cury
      ILI9325.draw(x+xoffset,y+yoffset,x+width-1+xoffset,y+height-1+yoffset)      ' draw on screen
      ILI9325.docmd("C",0,jump+32,size)              ' move bytes from ram out to the display  
    return x + xadvance
}
' hub version
PUB ILIChar(ascii,x,y) | jump,size,width,height,ramaddress,xoffset,yoffset,xadvance   ' read out a character from a .ifn file stored in ram
    jump := long[@font + (ascii << 2) +  256]' jump location = ascii *4 plus fonttable + 256
    jump += @font
    size := long[jump]                 ' size precalculated = width x height x 2
    xadvance :=  long[jump][5]     '     ' amount to move to next character offset = 20 which is 5 longs
    if size > 0 and ascii <> 32                                           ' no need to print anything if size is zero or read more font data
      width :=   long[jump][1]                ' width value of offset is 1 long
      height :=  long[jump][2]            ' height
      xoffset := long[jump][3]          ' xoffset to move
      yoffset := long[jump][4]           ' yoffset to move
      if (x+width -1 + xoffset) > screenwidth
        ILIcrlf ' do a new line if it won't fit
        x :=curx
        y :=cury
      ILI9325.draw(x+xoffset,y+yoffset,x+width-1+xoffset,y+height-1+yoffset)      ' draw on screen
      ILI9325.docmd("D",jump+32,0,size)                                 ' move bytes from ram out to the display   
    return x + xadvance

PUB ILIPrintstring(stringptr)                 'print at curx,cury
'' Print a zero-terminated string
  repeat strsize(stringptr)
    curx := ILIChar(byte[stringptr++],curx,cury)                           


PUB ILISetCursor(x,y)            
    curx := x
    cury := y

PUB ILIPrintstringcrlf(stringptr) ' print string and crlf
    ILIprintstring(stringptr)
    ILIcrlf
 
PUB ILIcrlf
    curx := 0
    cury += FontHeight
    ' don't clear background here, do in calling routine
    'ClearFontBackground(0,cury,screenwidth,cury + FontHeight) ' clear background line to the background color ready to draw          
{
PUB FullScreenRaw(stringptr) | ramaddress                         ' convert 3 byte .raw file to 2 byte .ili file
    fat.openfile(stringptr,"R")                ' open an image
    ramaddress := 0
    repeat 600
      fat.readdata(@sdbuffer,384)                       ' 128 pixels, 3 bytes per pixel
      ILI9325.docmd("E",@sdbuffer,@buffer2,128)         ' convert to 2 byte .ili format, len is number of pixels
      ILI9325.docmd("A",@buffer2,ramaddress,256)        ' A must start on 0,256,512,768 etc move bytes to external ram
      ramaddress := ramaddress + 256                    ' add 256
    fat.closefile
    RamToDisplay                              ' move bytes from ram out to display

PUB RamToDisplay ' dumps preloaded ram to display starting address 0
    'Clearscreen                                        ' leave out clearscreen
    ILI9325.draw(0,0,screenwidth,screenheight)                           ' set up the area of the screen to draw in
    ILI9325.docmd("C",0,0,153600)                       ' move bytes from ram out to the display

PUB Clearscreen(color) 
    bytefill(@sdbuffer,color,512)
    ILI9325.draw(0,0,screenwidth,screenheight)                           ' set up the area of the screen to draw in   
    repeat 300
      ILI9325.docmd("D",@sdbuffer,0,512)                       ' move bytes from ram out to the display
}
PUB ClearFontBackground(x1,y1,x2,y2) | i,size,remainder ' clears an area of the screen to the current font background color
    size := (x2-x1+1) * (y2 - y1 +1) ' number of pixels
    remainder := (size & 255) << 1                      ' if not a whole number of 256 pixels
    repeat i from 0 to 510 step 2                         ' move the background font colour to the buffer
      sdbuffer[i] := BackFontColor & 255                  ' and replicate for 256 pixels
      sdbuffer[i+1] := BackFontColor >> 8
    ILI9325.draw(x1,y1,x2,y2)                           ' set up the area of the screen to draw in
    repeat size >> 8                                    ' 512 byte blocks
      ILI9325.docmd("D",@sdbuffer,0,512)                ' 2 = 2 bytes which = one pixel
    if remainder <> 0  
      ILI9325.docmd("D",@sdbuffer,0,remainder)           ' do the remainder          


PUB NewFont(strpointer)
    ILIloadfont(strpointer)
    ' only call this routine with the cursor curx at zero
    ClearFontBackground(curx,cury,screenwidth,screenheight)             ' clear an area the same color as the font background  
    ILIsetcursor(0,0)
    
PUB Text_cls
   bytefill(@screen, " ", screensize)
   col := row := 0

PUB Text_crlf
   Text_out(13)
   Text_out(10)   

PUB Text_out(c) | i, k
'' Output a character
''
''     $08 = backspace
''     $09 = tab (8 spaces per)
''     $01 = set X position (X follows)
''     $0B = set Y position (Y follows)
''     $0C = set color (color follows)  (not on this simple terminal)
''     $0A = line feed
''     $0D = carriage return (col = 0)
''  others = printable characters

  case flag
    $00: case c
           $08: if col        ' not zero so can subtract one
                  col--
                  Text_print(" ")     ' backspace and rubout
                  col--
                  Text_RedrawLine
           $09: repeat
                  Text_print(" ")
                while col & 7
           $01: flag := c
             return
           $0B: flag := c
             return
           $0C: flag := c
             return
           $0A: Text_newline ' linefeed/new line     
           $0D: col := 0 ' carriage return
           other: Text_print(c)
    $0A: col := c // cols
    $0B: row := c // rows
    '$0C: color := c & 7
  flag := 0

PRI Text_print(c)         ' private routine from routine above
  screen[row * cols + col] := c  ' store in the screen text buffer
  ILIChar(c,col<<3,row <<4)   ' print on the screen  col*8 and row *16 as this is the font size
  if ++col == cols
    Text_newline

PRI Text_newline | i
  col := 0
  if ++row == rows
    row--
    bytemove(@screen, @screen[cols], lastrow)   'scroll lines
    bytefill(@screen[lastrow], " ", cols)      'clear new line
    Text_RedrawScreen                                   ' all lines shuffle up one so need to redraw all the screen

PUB Text_RedrawScreen | x,y,i
    ClearFontBackground(0,0,screenwidth,screenheight)
    ILIsetcursor(0,0)
    i := 0
    repeat y from 0 to rows -1
      repeat x from 0 to cols -1
        curx := ILIChar(screen[i],curx,cury)
        i++
      curx :=0
      cury +=fontheight

PRI Text_RedrawLine | oldcurx,i ' redraw current row, for backspace
   oldcurx := curx' store
   ClearFontBackground(0,row <<4 ,screenwidth,row <<4 + FontHeight) ' clear background line to the background color ready to draw
   curx := 0
   cury := row <<4 ' times 16
   i := row * cols ' variable row times constant cols
   repeat cols ' do 40x
     curx := ILIChar(screen[i],curx,cury)
     i++
   curx := oldcurx              ' restore curx

PUB Text_Str(stringptr)
'' Print a zero-terminated string
  repeat strsize(stringptr)
    Text_out(byte[stringptr++])
    
PUB Text_dec(value) | i
'' Print a decimal number
  if value < 0
    -value
    Text_out("-")
  i := 1_000_000_000
  repeat 10
    if value => i
      Text_out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      Text_out("0")
    i /= 10

PUB Text_hex(value, digits)
  {{ Send value as hexadecimal characters up to digits in length.
  ''
  '' `Parameters:
  ''
  ''     `value: byte, word, or long value to send as hexadecimal characters.
  ''     `digits: number of hexadecimal digits to send.  Will be zero padded if necessary.
  ''
  '' `Return: none.
  ''
  '' `Example: pst.Hex(1234, 5)
  ''
  ''     Output decimal 1234 as five hex digits. Outputs `004D2.
  ''     Next lines, if needed...
  }}

  if (digits > 8)
    repeat digits - 8
      Text_out("0")
    digits := 8
  else
    value <<= (8 - digits) << 2
  repeat digits
    Text_out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))                  
{
PUB PrintStringCR(pstring)' sends pstring to the vga and serial port
     PrintString(pstring)                               ' print the string  
     crlf

PUB PrintString(pstring)' sends pstring to the vga and serial port with no CRLF
    if (_TV_Enabled)
      text.str(pstring)
'    if (_VGA_Enabled)  
'      vt100.str(pstring)                                 ' send to vga screen
'    if (_Serial_Enabled)   
'    sio.str(0,pstring)                                 ' send to com port
'    if (_lcd_Enabled)
'       lcd.str(pstring)                                   ' send to 20x4 lCD display

PUB PrintChar(c) ' sends c to vga and serial port
    if (_TV_Enabled)
      text.out(c)
'    if (_VGA_Enabled)
'      vt100.out(c)                                        ' send to vga
'    if (_Serial_Enabled)
'      sio.tx(0,c)                                         ' send to com port
'    if (_lcd_Enabled)
'      LCD.out(c)                                          ' send to LCD display      
PUB CRLF
    PrintChar(13)                                        ' send to vga
    PrintChar(10)                                        ' send to vga

PUB PrintDecimal(number)                                ' print the decimal value, useful for debugging
     PrintString(str.integerToDecimal(number, 10))
     crlf

PUB PrintHex(number)
    Printstring(str.integertohexadecimal(number,8)) ' print the hex long value on the TV
}
DAT
sdbuffer        byte    $0[512]  ' 512 byte buffer for sd card interface
buffer2         byte    $0[512]  ' 512 general purpose hub buffer
font            byte    $0[13312]